/*
    Lovasz_ell_HR.cpp
    Supporting material software to the article 
    QUANTITATTIVE CONSTRAINT-BASED COMPUTATIONAL MODEL OF TUMOR-TO-STROMA COUPLING VIA LACTATE SHUTTLE
    
    Copyright (C) 2015 F.Capuani, D.De Martino, E.Marinari, A.De Martino

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
                                                                           */


/* 
                                            BRIEF DESCRIPTION
   INPUT:  The matrix constraint of the polytope (reduced row echelon form of the stoichiometric matrix), reaction bounds 
   NB the polytope should be full dimensional
   OUTPUT:   1) A rounding Lovasz Ellipsoid of the polytope (approximating the Lowner-John Ellipsoid)
   2) A sampling of the metabolic space according to the Boltzmann distribution  exp(BETA v)   BETA=level of maximization v=objective function (if BETA=0 Uniform sampling)
   obtained  with the Hit and Run algorithm  by using a rounding ellipsoid to avoid ill-conditioning
   NB The output of 2) are the independent variables in the polytope (fluxes can be obtained applying the matrix constraint) 
      1) is printed on a file while 2) is printed in the std output      
                                                                                                                                */

#include "iostream"
#include "stdlib.h"
#include "math.h"
#include "time.h"
#include "fstream"
#include "vector"
#include "string"
#include <gsl/gsl_math.h>			    // This software uses gnu scientific library (GSL) to implement matrix diagonalization.
#include <gsl/gsl_eigen.h>                          // We link the GSL libraries (on linux) with: g++ lovasz_ellispoid_HR.cpp `gsl-config --cflags --libs`
using namespace std;

#define r 34 		// number of independent variables
#define N 153 		// number of fluxes
#define SAMPLETIME 1000000 		// number of TOTAL SAMPLE POINTS
#define POINTSWANTED 1000 	// number of points to keep of the total SAMPLE points
#define BETA 50	// Lagrange multiplier of the objective function to maximize with H&R
#define ID_RXN_TO_MAX 74	// ID of the flux that will be maximized with BETA
#define POLYHEDRON_FNAME "HCCN_2CELLS_RREF.txt"	// Name of the file containing the polytope constraints
#define REACTIONS_BOUNDS_FNAME "HCCN_2CELLS_bounds.txt"	// Name of the file containing the bounds of the reactions
#define INPUTANN "ellipsoid_lovasz.txt"	// Filename where ellipsoid computed with Lovasz is written 
#define MAXFLUX 10000.000000	// Initialization of variables tp, tm in extrema(). MAXFLUX should be larger than maximum flux
#define EPSILON	1e-30                 
#define MAXIMUM_EXP 20
long long int samplerate;                  // sampling rate
std::vector<int> matrice[N];                     // constraint matrix pointers
std::vector<double> matrice2[N];                 // constraint matrix values 
double   boundmin[N];                            // reaction bounds: min flux
double   boundmax[N];                            // reaction bounds: max flux            
double   flux[r];                                // vector with the independent variables   
double   versori[r][r];                          // direction of the ellipsoid axis 
double   cb[r];                                  // vector encoding the objective function              
double   ran[2];                                 // random variables for gaussian generator
double   lenght[r];                              // ellipsoid axis lengths




void reading_lovasz(){                               // READING THE POLYTOPE
    fstream file;
    string nomi;
    file.open(REACTIONS_BOUNDS_FNAME, ios::in);

    for(int i=0;i<=N-1;i++){
	file >> nomi >> boundmin[i] >> boundmax[i];

    }                                          
    file.close();
    file.open(POLYHEDRON_FNAME, ios::in);                                     // polytope constraints reading
    int  i1=0;
    int  j1=0;        
    double pepe;
    while(file >> pepe){
	if(pepe!=0){
	    matrice[i1].push_back(j1);
	    matrice2[i1].push_back(pepe);
	}
	j1++;
	if(j1==r){
	    j1=0;
	    i1++;
	}
    }        
    file.close();   


    




}






void inizialization_lovasz(){                                      // INITIALIZATION OF LOVASZ ALGORITHM
     for(int i=0;i<=r-1;i++) flux[i]=0;
    
    for(int i=0;i<=r-1;i++) for(int j=0;j<=r-1;j++){
	if(i==j) versori[i][j]=1;
	else     versori[i][j]=0;
    }
    double maxi = 0;
    for(int i=0;i<=N-1;i++) if(boundmax[i]-boundmin[i]>maxi) maxi = boundmax[i]-boundmin[i];
    double Radius =sqrt(r)*maxi;

    for(int i=0;i<=r-1;i++) lenght[i] = Radius;



}



void basicellipsoid(int value){              // find a point inside with ellipsoidal technique, update step
    
                    
   
   
    
    double Matr[r][r];
    
    for(int i=0;i<=r-1;i++) for(int j=0;j<=r-1;j++) Matr[i][j]=0;
 
    for(int k=0;k<=r-1;k++) for(int i=0;i<=r-1;i++) for(int j=0;j<=r-1;j++) Matr[i][j] += versori[k][i]*versori[k][j]*lenght[k]*lenght[k]; 


    double fluxstep[r];
    double norm=0;
    for(int i=0;i<=r-1;i++) fluxstep[i]=0;
    int sign=1;
    if(value>=N) sign=-1;    

    if(sign==1){  
                  double newvec[r];
                  for(int i=0;i<=r-1;i++) newvec[i]=0;
                  for(unsigned int i=0;i<=matrice[value].size()-1;i++) newvec[matrice[value][i]]=-matrice2[value][i];
                  for(int i=0;i<=r-1;i++){
                                  double x=0;
                                  for(int j=0;j<=r-1;j++) x += Matr[i][j]*newvec[j];
                                  fluxstep[i]=x;
                                  norm += x*newvec[i];
                                  }         
                }
    else{ 
                  double newvec[r];
                  for(int i=0;i<=r-1;i++) newvec[i]=0;
                  for(unsigned int i=0;i<=matrice[value-N].size()-1;i++) newvec[matrice[value-N][i]]=matrice2[value-N][i];
                  for(int i=0;i<=r-1;i++){
                                  double x=0;
                                  for(int j=0;j<=r-1;j++) x += Matr[i][j]*newvec[j];
                                  fluxstep[i]=x;
                                  norm += x*newvec[i];
                                  }    
                  



                 }
   
    norm=sqrt(norm);
    for(int i=0;i<=r-1;i++) fluxstep[i]/=norm;
    double rr=r;
    double a1,a2,a3;
    a1 = 1./(rr+1.);
    
    for(int i=0;i<=r-1;i++) flux[i]   -=   a1*fluxstep[i];
                           
                        
    a2 = rr*rr/(rr*rr-1.);
    a3 = 2.;
       
   



    
    



    for(int i=0;i<=r-1;i++) for(int j=0;j<=r-1;j++) Matr[i][j] =    a2*(Matr[i][j]-a3*a1*fluxstep[i]*fluxstep[j]);   // assign matrix shallow cut

                                       
                                       
                                       
                                        
    double D[r];                                                                             
    double Matr2[r*r];
    for(int i=0;i<=r*r-1;i++) Matr2[i]=Matr[i/r][i%r];

    
 

  gsl_matrix_view matricia 
    = gsl_matrix_view_array (Matr2, r, r);

  gsl_vector *eval = gsl_vector_alloc (r);
  gsl_matrix *evec = gsl_matrix_alloc (r, r);

  gsl_eigen_symmv_workspace * workspace = 
    gsl_eigen_symmv_alloc (r);
   
  gsl_eigen_symmv (&matricia.matrix, eval, evec, workspace);

  gsl_eigen_symmv_free (workspace);

  gsl_eigen_symmv_sort (eval, evec, 
                        GSL_EIGEN_SORT_ABS_ASC);

  
  {
    int i;

    for (i = 0; i <= r-1; i++)
      {
        D[i] = gsl_vector_get (eval, i);
      
        for(int j=0;j<=r-1;j++) Matr[i][j]=gsl_matrix_get(evec,i,j);
      
      }
  }

  gsl_vector_free (eval);
  gsl_matrix_free (evec);







    for(int i=0;i<=r-1;i++){                                  
                                  if(D[i]<0) D[i]=-D[i];
                                  lenght[i]=sqrt(D[i]);                                
                          }
    for(int i=0;i<=r-1;i++) for(int j=0;j<=r-1;j++) versori[i][j]=Matr[j][i];
   
}

void newellipsoid(int value){                // ellipsoid Lovasz update


    double beta;
    int sign=1;
    if(value>=N) sign=-1;
    if(sign==1){
	beta=-boundmin[value];
	for(unsigned int i=0;i<=matrice[value].size()-1;i++) beta += matrice2[value][i]*flux[matrice[value][i]];
    }
    else{
	beta=boundmax[value-N];
	for(unsigned int i=0;i<=matrice[value-N].size()-1;i++) beta -= matrice2[value-N][i]*flux[matrice[value-N][i]];
    }


    double Matr[r][r];


    for(int i=0;i<=r-1;i++) for(int j=0;j<=r-1;j++) Matr[i][j]=0;

    for(int k=0;k<=r-1;k++) for(int i=0;i<=r-1;i++) for(int j=0;j<=r-1;j++) Matr[i][j] += versori[k][i]*versori[k][j]*lenght[k]*lenght[k]; 


    double fluxstep[r];
    double norm=0;
    for(int i=0;i<=r-1;i++) fluxstep[i]=0;

    if(sign==1){  
	double newvec[r];
	for(int i=0;i<=r-1;i++) newvec[i]=0;
	for(unsigned int i=0;i<=matrice[value].size()-1;i++) newvec[matrice[value][i]]=-matrice2[value][i];
	for(int i=0;i<=r-1;i++){
	    double x=0;
	    for(int j=0;j<=r-1;j++) x += Matr[i][j]*newvec[j];
	    fluxstep[i]=x;
	    norm += x*newvec[i];
	}         
    }
    else{ 
	double newvec[r];
	for(int i=0;i<=r-1;i++) newvec[i]=0;
	for(unsigned int i=0;i<=matrice[value-N].size()-1;i++) newvec[matrice[value-N][i]]=matrice2[value-N][i];
	for(int i=0;i<=r-1;i++){
	    double x=0;
	    for(int j=0;j<=r-1;j++) x += Matr[i][j]*newvec[j];
	    fluxstep[i]=x;
	    norm += x*newvec[i];
	}    




    }

    norm=sqrt(norm);
    beta/=norm;
    double rr=r;
    double a1,a2,a3;
    a1 = (1.-rr*beta)/(rr+1.);

    for(int i=0;i<=r-1;i++){

	fluxstep[i]/=norm;
	flux[i]   -=   a1*fluxstep[i];

    }
    a2 = rr*rr*(1.-beta*beta)/(rr*rr-1.);
    a3 = 2./(1.-beta);
 for(int i=0;i<=r-1;i++) for(int j=0;j<=r-1;j++){


	Matr[i][j] =    a2*(Matr[i][j]-a3*a1*fluxstep[i]*fluxstep[j]);   // assign matrix shallow cut

    }
   

    

  double D[r];                                                                             
    double Matr2[r*r];
    for(int i=0;i<=r*r-1;i++) Matr2[i]=Matr[i/r][i%r];

    


  gsl_matrix_view matricia 
    = gsl_matrix_view_array (Matr2, r, r);

  gsl_vector *eval = gsl_vector_alloc (r);
  gsl_matrix *evec = gsl_matrix_alloc (r, r);

  gsl_eigen_symmv_workspace * workspace = 
    gsl_eigen_symmv_alloc (r);
  
  gsl_eigen_symmv (&matricia.matrix, eval, evec, workspace);

  gsl_eigen_symmv_free (workspace);

  gsl_eigen_symmv_sort (eval, evec, 
                        GSL_EIGEN_SORT_ABS_ASC);
  
  {
    int i;

    for (i = 0; i <= r-1; i++)
      {
        D[i] = gsl_vector_get (eval, i);
      
        for(int j=0;j<=r-1;j++) Matr[i][j]=gsl_matrix_get(evec,i,j);
      
      }
  }

  gsl_vector_free (eval);
  gsl_matrix_free (evec);








   


  


    for(int i=0;i<=r-1;i++){    
	
	if(D[i]<0) D[i]=-D[i];
	lenght[i] = sqrt(D[i]);

	

    }
    for(int i=0;i<=r-1;i++) for(int j=0;j<=r-1;j++) versori[i][j]=Matr[j][i];
   
}

int checkinside(){                 // check if the point is inside
    int check=-1;
    int r1=0;
    do{         

	int min,sign;
	double xmin=10000;
	int i=0;  
	do{
	    if(matrice[i].size()>0){
		double x=0;
		for(unsigned int j=0;j<=matrice[i].size()-1;j++) x += flux[matrice[i][j]]*matrice2[i][j];
		double x1 = x-boundmin[i];
		double x2 = boundmax[i]-x;
		if(x1<xmin){
		    xmin = x1;
		    min =  i;
		    sign = 1;
		}
		if(x2<xmin){
		    xmin = x2;
		    min =  i;
		    sign = -1;
		}

	    }
	    i++;
	}while(i<N && xmin>=0);
	if(xmin<0){
	    check =  min;
	    if(sign<0) check+=N;   

	}
	else r1++;

    }while(r1<r && check<0);        


    return check;

}



int checkaxis(){               // check if the ellipsoid gives a good rounding
    int check=-1;
    int r1=0;   
    double div = r+1;
    do{         
	double punto1[r];
	double punto2[r];
	for(int i=0;i<=r-1;i++){
	    punto1[i]=punto2[i]=flux[i];
	    double step = lenght[r1]*versori[r1][i]/div;
	    punto1[i]+=step;
	    punto2[i]-=step;
	}   
	int min,sign;
	double xmin=10000;  
	int i=0;
	do{
	    if(matrice[i].size()>0){
		double x=0;
		for(unsigned int j=0;j<=matrice[i].size()-1;j++) x += punto1[matrice[i][j]]*matrice2[i][j];
		double x1 = x-boundmin[i];
		double x2 = boundmax[i]-x;
		if(x1<xmin){
		    xmin = x1;
		    min =  i;
		    sign = 1;
		}
		if(x2<xmin){
		    xmin = x2;
		    min =  i;
		    sign = -1;
		}
		x=0;
		for(unsigned int j=0;j<=matrice[i].size()-1;j++) x += punto2[matrice[i][j]]*matrice2[i][j];
		x1 = x-boundmin[i];
		x2 = boundmax[i]-x;
		if(x1<xmin){
		    xmin = x1;
		    min =  i;
		    sign = 1;
		}
		if(x2<xmin){
		    xmin = x2;
		    min =  i;
		    sign = -1;
		}
	    }
	    i++;
	}while(i<N && xmin>=0);
	if(xmin<0){
	    check =  min;
	    if(sign<0) check += N;   

	}
	else r1++;

    }while(r1<r && check<0);        


    return check;

}








double casual(){                                                // Random number uniform in (0,1)
    double   x = double  (random())/(RAND_MAX+1.0);
    return x;
}



void gaussMP(double& a,double& b){                  // Couple of Gaussian random numbers
    double   x,y,s;
    do{
	x=-1+2*casual();
	y=-1+2*casual();
	s = x*x+y*y;
    }while(s>=1);
    double   c = sqrt(-2*log(s)/s);
    a = x*c;
    b = y*c;
}





void lovasz_ellipsoid(){              // building Lovasz ellipsoid


    reading_lovasz();                  // reading constraints
    inizialization_lovasz();              // initialization
    int check; 
   


    do{
	check=1000;
	int check0 = checkinside();            // check if the point is inside
	if(check0<0){                       // if the point is inside find the rounding ellipsoid
	    check =  checkaxis();             // check if ellipsoid gives a good rounding

	    if(check>=0) newellipsoid(check);

	}
	else{                              // else find a point inside
	    basicellipsoid(check0);



	}
    }while(check>=0);
    
    fstream file;                                          // write the ellipsoid in a file
    file.open(INPUTANN , ios::out);
    for(int i=0;i<=r-1;i++) file << flux[i] << "  ";
    file << endl;
    for(int i=0;i<=r-1;i++){
	file << lenght[i] << endl;
	for(int j=0;j<=r-1;j++) file << versori[i][j] << "  ";
	file << endl;
    }           
    file.close();
   

}


void initHR(){                               // Hit and Run initialization 
    samplerate=1;
    if (POINTSWANTED<SAMPLETIME) samplerate = ( (double) SAMPLETIME)/( (double) POINTSWANTED);   // sampling rate
    for(int i=0;i<=r-1;i++){                                // Objective function
        cb[i]=0;
        for(unsigned int j=0;j<=matrice[ID_RXN_TO_MAX].size()-1;j++)  cb[i] += versori[i][matrice[ID_RXN_TO_MAX][j]]*matrice2[ID_RXN_TO_MAX][j];
    }
}





void findextrema(double& tp, double& tm, double n[]){          // given direction n, find the extremes of segment passing from current point within the Polytope
    tp=MAXFLUX;
    tm=-MAXFLUX;
    for(int i=0;i<=N-1;i++){
       if(matrice[i].size()>0){
	double x,y;
	x=y=0;
	for(unsigned int j=0;j<=matrice[i].size()-1;j++){
	    x += flux[matrice[i][j]]*matrice2[i][j];
	    y += n[matrice[i][j]]*matrice2[i][j];
	}
	double   t1,t2;
	t1 = (boundmin[i]-x)/y;
	t2 = (boundmax[i]-x)/y; 
	if(t1<0 && t1>tm) tm=t1;
	if(t2<0 && t2>tm) tm=t2;
	if(t1>0 && t1<tp) tp=t1;
	if(t2>0 && t2<tp) tp=t2;
      }
    }
}



void marsagliaAC(double n[]){                       // extract uniformly  at random a point on the ellipsoid surface
    double norm=0;
    double n1[r];
    for(int i=0;i<=r-1;i++){  
	if(i%2==0) gaussMP(ran[0],ran[1]);                  // find a point on the r-sphere by Marsaglia & transform it with the ellipsoid matrix
	n1[i]=ran[i%2]*lenght[i];
    }
    for(int i=0;i<=r-1;i++){
	n[i]=0;
	for(int j=0;j<=r-1;j++) n[i] += n1[j]*versori[j][i];
	norm+=n[i]*n[i];
    }
    norm = sqrt(norm);
    for(int i=0;i<=r-1;i++) n[i]/=norm;
}

void ellipsHR(){                       // single Hit and Run Step
    double  n[r];
    marsagliaAC(n);                   // extract uniformly  at random a point on the ellipsoid surface
    double   tp,tm;
    findextrema(tp,tm,n);             // given direction n, find extremes of segment passing from current point
    double t;
     do{ 

      double ranvar = casual();
      if(BETA==0) t = tm+(tp-tm)*ranvar;                         // extract a random point on the segment (uniformly at random)
      else{
	double c=0;
	for(unsigned int j=0;j<=matrice[ID_RXN_TO_MAX].size()-1;j++)  c += n[matrice[ID_RXN_TO_MAX][j]]*matrice2[ID_RXN_TO_MAX][j];
	c*=BETA;

	if(fabs(c*(tp-tm))<MAXIMUM_EXP) t = tm + log( 1 + ranvar*(exp(c*(tp-tm))-1))/c;                  // extract a random point on the segment (according to BOLTZMANN) 
        else{
                    if(c<0) t = tm+log(1-ranvar)/c;
                    else    t = tp+log(ranvar)/c;
             }
        
       }
        
    }while(t-tm<EPSILON || tp-t<EPSILON);
    for(int i=0;i<=r-1;i++) flux[i] += t*n[i];
}











void Lov_Ell_Hit_and_run(){
     initHR();                                 // initialization
     for(int i=0;i<=SAMPLETIME;i++){                   // repeated Hit_and_Run steps
	ellipsHR();
	if(i%samplerate==0){                                           
	    for(int i=0;i<=r-1;i++) cout << flux[i] << "  ";             // write to std output independent variables (fluxes can be obtained applying the matrix constraint)  
	    cout << endl;
            }
        }   
}




int main(){
    cout.precision(16);
    srand(time(0));              // seed random number generator
    lovasz_ellipsoid();          // build the Lovasz ellipsoid               
    Lov_Ell_Hit_and_run();        // perform the Hit and Run 
    return 0;
}



